/*
--------------------------------------------------------------------------------
 Symbra - a symbol ranking file (de-)compressor
 (c) 2007, Frank Schwellinger, frank.schwellinger@gmx.de
 to compile: g++ -O2 -march=pentiumpro -Wall -s -o symbra symbra-0.2.cpp
--------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version. This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
details at http://www.gnu.org/copyleft/gpl.html.

--------------------------------------------------------------------------------
Program description
--------------------------------------------------------------------------------
Symbra is a file (de-)compressor similar to Peter Fenwicks symbol ranking 
algorithm 'srank', available at:

ftp://ftp.cs.auckland.ac.nz/pub/staff/peter-f/srank.c

The srank algorithm, designed for speed rather than good compression, is a
4-way set associative cache with a LRU update in order-3 contexts. When a
symbol is encountered it is encoded with 1, 3, or 4 bits according to its
rank in a queue of length 3, then moved to the front. Long runs of rank-0
are run length encoded using 12 bits to encode the length of the run. A miss
is coded using pseudo-MTF (swapping the symbol with another symbol about half
way to the front) in an order-0 context using 7 bits for the first 32 symbols
and 12 bits for the rest.

In contrast to the recently released C++ versions sr2 by Matt Mahoney and sr3 by
Francesco Antonio Nania (see http://cs.fit.edu/~mmahoney/compression/text.html),
which are modernized and improved versions of srank, Symbra follows more the
classic approach of srank: e.g. in that it uses no arithmetic coding. While
being more similar to srank than sr2 or sr3, Symbra still differs substantially
from srank.

Symbra allows contexts from order-0 to order-6 via commandline argument and
uses by default an order-4 context. The contexts are packed into 3 Bytes.
For the choosen context length a queue of length 3 with LRU update is
maintained. First Symbra makes internal static code assignments similar to
srank: an encountered symbol is assigned to 1, 2, or 3 bits according to its
queue position, then moved to the front. Runs of 9 rank-0s are assigned to a
5 bit code. A miss gets assigned to an 8 or 13 bit code determined by the
symbols MTF-index in an order-2 context, after that the symbol is moved (not
swapped) about the half way to the front.

--------------------------------------------------
Static code assignments
--------------------------------------------------
Code            Length    Description
--------------  --------  -------------------------
0                 1 bit   order-4 context rank-0
10                2 bits  order-4 context rank-1
110               3 bits  order-4 context rank-2
1110 xxxx       4+4 bits  order-2 MTF-index <16
11110 xxxxxxxx  5+8 bits  order-2 MTF-index >=16
11110 00000000  5+8 bits  EOB indicator
11111             5 bits  Run of 9 rank-0s
--------------------------------------------------

For the final encoding Symbra uses adaptive codes. That means the above listed
code assignments are encoded by their adaptively updated frequencies in an
order-2 context. The most frequent static code now gets encoded with the
shortest code and the least frequent static code gets encoded with the longest
code.

-----------------------------------------------------
Adaptive code assignments
-----------------------------------------------------
Code   Length   Description
-----  -------  -------------------------------------
0       1 bit   static-code order-2-frequency rank-0
10      2 bits  static-code order-2-frequency rank-1
110     3 bits  static-code order-2-frequency rank-2
1110    4 bits  static-code order-2-frequency rank-3
11110   5 bits  static-code order-2-frequency rank-4
11111   5 bits  static-code order-2-frequency rank-5

(eventually followed by a 4 or 8 bit order-2
MTF-index, or an 8 bit EOB marker)
-----------------------------------------------------

Symbra pushes the generated codes and MTF-indices and EOB marks into a 64 bit
FIFO queue, from where they are bytewise pulled out into a blockbuffer, while a
second blockbuffer gets filled with the original uncompressed bytes. Symbra
manages the output in small blocks of variable length to reduce the expansion
of data that e.g. srank produces when it processes precompressed data. If the
compression of a block fails, the blockbuffer with the original uncompressed
data is used for that block instead, thus exchanging sranks expansion
(<= ~40.9%) with a much smaller one (<= ~1.3%), which is introduced by the block
structures and by specific codes needed for the block management: e.g the first
byte of a block, which is used as the block header indicating the type (1 bit)
and the length (7 bits) of a block, and the (adpative) EOB marker (<= 13 bits)
which indicates the end of a block. The last block is always transmitted
uncompressed with a 3 byte block header.

--------------------------------------------------------------------------
Block Structures
--------------------------------------------------------------------------
Bit 1  Type          Bit 2-8              followed by
-----  ------------  -------------------  --------------------------------
1      compressed    7 bits block length  block data + EOB
0      uncompressed  7 bits block length  block data
0      uncompressed  7 zeroes=last block  16 bit block length + block data
--------------------------------------------------------------------------

When in encoding mode Symbra can optionally process the input in 2 passes. If
choosen so, it maintains a table with order-0 symbol frequencies in the first
pass. In the second pass that table is used for initialization of the
context-queues and the order-2 MTF-tables as follows: every context-queue is
initialized with the 3 most frequent order-0 symbols and every MTF-table is
initialized with the 256 symbols sorted by their order-0 frequency. The 256
frequency-sorted symbols are written to the header of the output file. The
decoder loads the symbols from the header and initializes its tables according
to the encoder, thus decoding always takes only 1 pass regardless whether
the source was encoded in 1 or 2 passes. The 2-pass mode improves compression
only by a small degree and increases the encoding time significantly, thus
Symbra encodes by default in 1-pass, initializing the context-queues with 0
values and the mtf-tables with values from 0 to 255.

-----------------------------------------------------------------------------
File Header Structure
-----------------------------------------------------------------------------
Bytes  Description
-----  ----------------------------------------------------------------------
    1  Magic Byte
    1  Program Version
    1  Program Subversion
    1  Contextmode (0-6)
    1  Matchmode (0-5)
    1  Passmode (1 or 2)
  256  Symbols sorted by frequency (most frequent first) [only if Passmode=2]
-----------------------------------------------------------------------------

Optionally Symbra allows to merge the context queues with the results of suffix
match searches. If choosen so, it stores the symbol following the longest
suffix match at the top of the context queue. Activating the suffix match search
slows down the processing time considerable, but it shows that Symbras
compression rates can get close to those of LZ algorithms or even surpass them
in some situations without using an entropy coder. The suffix match search is
implemented as an unbalanced binary search tree, which is too slow for practical
use (at least the way it is implemented), but it is designed to allow searches
in buffers up to a size of 455 MB, assumed the appropriate free memory (buffer
size x 9) is available. In this version of Symbra the buffer size can be
configured with a command line argument up to a length of 5 MB in steps of 1 MB
and the suffix search depth is also tied to the choosen buffer length, allowing
maximum suffix match lengths from a maximum matchlength of 4 for a 1 MB buffer
up to a maximum matchlength of 64 for a 5 MB buffer.

This version of Symbra uses 3x16 MB memory for the context-queues and 16 MB
memory for the order-2 MTF-table (an option would be to implement hashes 
therefor to reduce memory usage and memory access times). In addition Symbra
uses 9 MB memory for every 1 MB of search buffer requested by commandline
argument, and it uses a few kbytes memory for other variables and for the
program code.

------------------------------------------------------------------
Memory Usage
------------------------------------------------------------------
Usage    Description
-------  ---------------------------------------------------------
  48 MB  context queues: 16 MB x length of queue(=3)
  16 MB  order-2 MTF tables
0-45 MB  binary search tree for suffix matching: buffer length x 9
  xx KB  other variables and program code
------------------------------------------------------------------

Listed below is a comparison of some compression results of the 3 symbol ranking
algorithms srank, Symbra and sr2. All 3 programs were compiled with g++ 4.1.1
and tested on a Pentium III under Linux. Execution times are listed as reported
by the programs and times are not averaged. Options used were: -C8 for srank and
none for Symbra and sr2. Compressed size is listed as reported by the programs.
The fastest algorithm is srank, the best compression has sr2. Symbra lies in
between with a better compression than srank and execution times close to those
of sr2.

-------------------------------------------------------------------
Comparison of symbol ranking compressors
-------------------------------------------------------------------
                     compressed size in bytes / time in seconds
File          srank -C8         Symbra            sr2
------------  ---------- ------ ---------- ------ ---------- ------
FP.LOG         2.809.979   3.20  1.331.054   6.00    699.171   5.35
world95.txt    1.287.779   0.71  1.073.022   2.27    796.777   1.95
english.dic    1.005.306   0.72    873.833   2.02    649.876   1.81
vcfiu.hlp      1.382.532   0.86  1.175.563   2.48    882.891   2.09
ohs.doc        1.607.891   0.94  1.183.162   3.63    953.569   2.43
AcroRd32.exe   2.263.823   1.06  2.161.768   4.23  1.903.791   3.78
MSO97.DLL      2.770.218   1.20  2.588.800   4.91  2.274.079   4.49
rafale.bmp     1.690.168   0.90  1.352.229   2.44  1.006.034   2.24
A10.jpg        1.168.949   0.40    850.458   2.98    841.840   1.58
FlashMX.pdf    5.237.480   1.92  3.888.676  10.87  3.850.838   7.17
gimp.tar      23.959.818  16.26 19.579.002  36.41 15.321.253  38.86 
enwik8        43.091.429  23.95 38.308.163  53.26 30.432.506  67.58
hsp.mp3       26.864.064   9.51 19.776.572  51.98 19.285.671  35.51
dosgames.tar  42.284.085  20.95 33.920.399  71.15 32.706.311  66.13
zeroes.txt            14   0.61     77.543   1.41         23   0.92
kdelibs.tar   41.370.978  23.06 38.272.781  62.76 32.787.568  67.95
maxicomp.tar  21.289.404  12.02 16.418.749  35.52 14.233.695  32.72
enwik8.xwrt   33.369.954  14.77 29.410.247  46.17 26.621.114  52.76
-------------------------------------------------------------------

The table below shows the same files, but this time Symbras results
are obtained with the command line arguments giving the best
compression result for the appropriate file (i have not tested all
possible combinations though). Gzip 1.3.5 with option -9 represents
the LZ77 algorithm. Execution times are not listed, but Symbra gets
about 20x slower if suffix matching (-m) is activated. While Symbra
is no competitor for the other two programs in terms of execution
times, it shows that it is sometimes possible to achieve comparable
compression rates with the classic approach, in some cases even
surpassing LZ77 and the binary arithmetic coder of sr2 and sr3.

-------------------------------------------------------------
Comparison of compressors
-------------------------------------------------------------
                  compressed size in bytes
File          gzip -9    Symbra     sr2        Symbra options
------------  ---------- ---------- ---------- --------------
FP.LOG         1.337.914  1.256.767    699.171    -c6 -p2
world95.txt      862.955    747.674    796.777    -c5 -p2 -m5
english.dic    1.050.495    853.060    649.876        -p2
vcfiu.hlp        840.488    861.788    882.891    -c5 -p2 -m5
ohs.doc        1.007.497    985.862    953.569    -c5 -p2 -m5
AcroRd32.exe   1.732.385  1.956.179  1.903.791    -c2 -p2 -m5
MSO97.DLL      2.188.891  2.328.535  2.274.079    -c2 -p2 -m5
rafale.bmp     1.254.355  1.193.955  1.006.034    -c3 -p2 -m5
A10.jpg          841.463    850.442    841.840    -c2     -m5
FlashMX.pdf    3.833.206  3.820.075  3.850.838        -p2 -m5
gimp.tar      18.251.434 13.097.680 15.321.253    -c6 -p2 -m5 
enwik8        36.445.248 34.683.661 30.432.506    -c5 -p2 -m5
hsp.mp3       19.401.068 19.775.243 19.285.671    -c2 -p2 -m5
dosgames.tar  32.036.441 30.215.857 32.706.311        -p2 -m5
zeroes.txt         4.784     77.116         23    -c0
kdelibs.tar   33.354.899 31.655.075 32.787.568    -c6 -p2 -m5
maxicomp.tar  14.958.474 15.391.070 14.233.695        -p2 -m5
enwik8.xwrt   26.912.498 27.766.892 26.621.114    -c3 -p2 -m5
-------------------------------------------------------------

Conclusion: The most time consuming part of this Symbra version is
the suffix match search. With a faster suffix match implementation
or a better solution as substitution for it, it should be possible
to design a practical coder with moderate memory requirements
following the above described approach.

Note: Only input files with a file size up to 4 GB are supported
by this version of Symbra (this is not tested by the program).

--------------------------------------------------------------------------------
HISTORY
--------------------------------------------------------------------------------
version  from date   until date   changes
-------  ----------- -----------  ----------------------------------------------
0.1      16-Nov-2007 26-Nov-2007  initial version
0.2      27-Nov-2007 29-Nov-2007  added suffix match search; small modifications
--------------------------------------------------------------------------------
*/
//------------------------------------------------------------------------------
// INCLUDED HEADER FILES
//------------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
//------------------------------------------------------------------------------
// OTHER COMPILER DIRECTIVES
//------------------------------------------------------------------------------
//#define VERBOSEMODE 1 //for testing
#define UINT8 unsigned char //unsigned 8bit integer
#define UINT32 unsigned int //unsigned 32bit integer
#define SINT32 int //signed 32bit integer
#define UINT64 unsigned long long int //unsigned 64bit integer
//------------------------------------------------------------------------------
// GLOBALS
//------------------------------------------------------------------------------
//---Program Version
const UINT8 VERSION=0;
const UINT8 SUBVERSION=2;
//---Program modes
UINT8 contextmode;
UINT8 passmode;
UINT8 matchmode;
//---Generated file header
UINT32 headerlength;
const UINT8 MAGICBYTE=233; //identifies generated stream
//---LRU Cache
UINT8 cachearray[256*256*256*3];
UINT32 cachectx=0;
UINT32 cacheoffset=0;
UINT32 rank0=0,rank1=0,rank2=0;
//---Order-2 Move-to-front table
UINT8 mtf[256*256][256];
UINT32 mtfpos=0;
UINT32 mtfctx=0;
UINT8 mtftemp=0;
UINT32 mtftarget=0;
//---Symbol Frequencies
typedef struct fr
{
	UINT8 character;
	UINT32 count;
}frqtype;
frqtype freq[256];
//---Number of code assignments
const UINT32 NCODES=6;
//---Code Frequencies
typedef struct co
{
	UINT8 code;
	UINT32 count;
}codetype;
codetype codefreq[256][NCODES];
UINT32 cfctx=0;
//---64bit FIFO queue
SINT32 bits=0;
UINT64 q=0;
const SINT32 QLEN=(sizeof(q)*8);
const UINT64 QMASK=0xFFFFFFFFFFFFFFFFULL;
const UINT32 RUNLENGTHLIMIT=9; //above this value FIFO queue may overflow
//---Block buffers
const SINT32 BLOCKLIMIT=127;
UINT8 outbuf[(BLOCKLIMIT*8*RUNLENGTHLIMIT)+13];
UINT8 orgbuf[(BLOCKLIMIT*8*RUNLENGTHLIMIT)+13];
UINT32 orgbufptr=0;
//---other globals used by encoder and decoder
UINT32 ch=0;
UINT32 outbytes=0, inbytes=0;
double totaltime;
FILE *in,*out;
//---Match searching
UINT32 matchlen=0;
UINT32 inbufsize=0; 
UINT8 *inbuf; //input buffer (dynamic allocation at runtime)
UINT32 *bstrightchild; //binary search tree right child nodes (dynamic allocation at runtime)
UINT32 *bstleftchild; //binary search tree left child nodes (dynamic allocation at runtime)
UINT32 inbufprefix=65; //Size in bytes of input buffer prefix
UINT32 inbufpos=65; //(Start) position in the input buffer
UINT8 matchchar=0; //char following longest match
UINT8 tmpmatchchar=0;
UINT32 bstnode=0;
UINT32 bstparent=0;
UINT32 bstcomparison=0;
UINT32 bstlimit=0;
//------------------------------------------------------------------------------
// RESET CODE FREQUENCIES
//------------------------------------------------------------------------------
inline void resetcodefreq()
{
	UINT32 i,ii;
	//---reset code frequencies
	for(ii=0;ii<256;ii++) {
		for(i=0;i<NCODES;i++) {
			codefreq[ii][i].code=(UINT8)i;
			codefreq[ii][i].count=0;
		}
	}
	cfctx=0;
}
//------------------------------------------------------------------------------
// INIT MEMORY
//------------------------------------------------------------------------------
inline void initmemory()
{
	UINT32 i,ii;
	if(passmode==2){
		//---Init cache
		for(i=0;i<256*256*256*3;i++) {
			cachearray[i]=freq[i%3].character;
		}
		//---Init MTF
		for(ii=0;ii<256*256;ii++) {
			for(i=0;i<256;i++) {
				mtf[ii][i]=freq[i].character;
			}
		}
	}else{
		//---Init cache
		for(i=0;i<256*256*256*3;i++) {
			cachearray[i]=0;
		}
		//---Init MTF
		for(ii=0;ii<256*256;ii++) {
			for(i=0;i<256;i++) {
				mtf[ii][i]=i;
			}
		}
	}

	//---Init code frequencies
	resetcodefreq();

	switch (matchmode) {
		case 0: inbufsize=0; bstlimit=0; break;
		case 1: inbufsize=1024*1024; bstlimit=4; break;
		case 2: inbufsize=2*1024*1024; bstlimit=8; break;
		case 3: inbufsize=3*1024*1024; bstlimit=16; break;
		case 4: inbufsize=4*1024*1024; bstlimit=32; break;
		case 5: inbufsize=5*1024*1024; bstlimit=64; break;
	}

	if(matchmode>0){
		//---Init Input Buffer
		inbuf=(UINT8 *)malloc(sizeof(UINT8)*(inbufsize+inbufprefix));
		if (inbuf==NULL) {
			printf("Memory Allocation failed.\n");
			exit(1);
		}
		for(i=0;i<inbufsize+inbufprefix;i++) {
			inbuf[i]=i%256;
		}
		
		//---Init binary search tree
		bstrightchild=(UINT32 *)malloc(sizeof(UINT32)*(inbufsize+inbufprefix));
		if (bstrightchild==NULL) {
			printf("Memory Allocation failed.\n");
			exit(1);
		}
		bstleftchild=(UINT32 *)malloc(sizeof(UINT32)*(inbufsize+inbufprefix));
		if (bstleftchild==NULL) {
			printf("Memory Allocation failed.\n");
			exit(1);
		}
		for(i=0;i<inbufsize+inbufprefix;i++) {
			bstrightchild[i]=0;
			bstleftchild[i]=0;
		}
	}

}
//------------------------------------------------------------------------------
// REBUILD BINARY SEARCH TREE
//------------------------------------------------------------------------------
inline void rebuildbst()
{
	UINT32 i;
#ifdef verbosemode
	printf("\nrebuilding binary search tree...\n");
#endif
	//---reset binary search tree
	for(i=0;i<inbufsize+inbufprefix;i++) {
		bstrightchild[i]=0;
		bstleftchild[i]=0;
	}
	//---partially rebuild binary search tree
	for(i=(inbufsize/2)+inbufprefix;i<inbufprefix+inbufsize-1;i++) {
		//---update input buffer
		inbufpos++;
		inbuf[inbufpos]=inbuf[i];
		bstnode=inbufprefix;
		while(1) {
			if(bstnode==0) {
				bstcomparison=0;
				while(1) {
					if(inbuf[inbufpos-bstcomparison]>
					        inbuf[bstparent-bstcomparison]) {
						bstrightchild[bstparent]=inbufpos;
						break;
					}
					if(inbuf[inbufpos-bstcomparison]<
					        inbuf[bstparent-bstcomparison]) {
						bstleftchild[bstparent]=inbufpos;
						break;
					}
					bstcomparison++;
					if(bstcomparison>bstlimit) {
						break;
					}
				}
				break;
			} else {
				bstcomparison=0;
				while(1) {
					if(inbuf[inbufpos-bstcomparison]>
					        inbuf[bstnode-bstcomparison]) {
						bstparent=bstnode;
						bstnode=bstrightchild[bstnode];
						break;
					}
					if(inbuf[inbufpos-bstcomparison]<
					        inbuf[bstnode-bstcomparison]) {
						bstparent=bstnode;
						bstnode=bstleftchild[bstnode];
						break;
					}
					bstcomparison++;
					if(bstcomparison>bstlimit) {
						bstparent=bstnode;
						bstnode=0;
						break;
					}
				}
			}
		}
	}
}
//------------------------------------------------------------------------------
// UPDATE BINARY SEARCH TREE
//------------------------------------------------------------------------------
inline void updatebst()
{
	//---update input buffer
	inbufpos++;
	if(inbufpos==(UINT32)(inbufsize+inbufprefix)) {
		inbufpos=inbufprefix;
		rebuildbst();
	}
	inbuf[inbufpos]=ch;

	//---update binary search tree and search longest suffix-match
	matchlen=0;
	tmpmatchchar=0;
	matchchar=0;
	bstnode=inbufprefix;
	while(1) {
		if(bstnode==0) {
			bstcomparison=0;
			while(1) {
				if(inbuf[inbufpos-bstcomparison]>
					inbuf[bstparent-bstcomparison]) {
					bstrightchild[bstparent]=inbufpos;
					break;
				}
				if(inbuf[inbufpos-bstcomparison]<
					inbuf[bstparent-bstcomparison]) {
					bstleftchild[bstparent]=inbufpos;
					break;
				}
				bstcomparison++;
				if(bstcomparison>bstlimit) {
					break;
				}
			}
			break;
		} else {
			if(bstnode+1<(UINT32)(inbufprefix+inbufsize))
				tmpmatchchar=inbuf[bstnode+1];
			bstcomparison=0;
			while(1) {
				if(inbuf[inbufpos-bstcomparison]>
					inbuf[bstnode-bstcomparison]) {
					bstparent=bstnode;
					bstnode=bstrightchild[bstnode];
					break;
				}
				if(inbuf[inbufpos-bstcomparison]<
					inbuf[bstnode-bstcomparison]) {
					bstparent=bstnode;
					bstnode=bstleftchild[bstnode];
					break;
				}
				bstcomparison++;
				if(bstcomparison>=matchlen) {
					//---longest match
					matchlen=bstcomparison;
					matchchar=tmpmatchchar;
				}
				if(bstcomparison>bstlimit) {
					bstparent=bstnode;
					bstnode=0;
					break;
				}
			}
		}
	}
	//---end of bst
}
//------------------------------------------------------------------------------
// WRITE HEADER
//------------------------------------------------------------------------------
inline void writeheader(FILE *out)
{
	UINT32 i;
	fputc(MAGICBYTE,out);
	fputc(VERSION,out);
	fputc(SUBVERSION,out);
	fputc(contextmode,out);
	fputc(matchmode,out);
	fputc(passmode,out);
	headerlength=6;
	if(passmode==2){
		for(i=0;i<256;i++) {
			fputc(freq[i].character,out);
		}
		headerlength+=256;
	}
}
//------------------------------------------------------------------------------
// READ HEADER
//------------------------------------------------------------------------------
inline SINT32 readheader(FILE *in)
{
	UINT32 i;
	if(fgetc(in)!=MAGICBYTE)return 1;
	if(fgetc(in)!=VERSION)return 1;
	if(fgetc(in)!=SUBVERSION)return 1;
	contextmode=fgetc(in);if(contextmode>6)return 1;
	matchmode=fgetc(in);if(matchmode>5)return 1;
	passmode=fgetc(in);if(passmode==0 || passmode>2)return 1;
	headerlength=6;
	if(passmode==2){
		for(i=0;i<256;i++) {
			freq[i].character=fgetc(in);
		}
		headerlength+=256;
	}
	return 0;
}
//------------------------------------------------------------------------------
// UPDATE CODE FREQUENCIES
//------------------------------------------------------------------------------
inline void updatecodefreq(UINT32 code)
{
	UINT32 i=0;
	UINT8 tmpcode=0;
	UINT32 tmpcount=0;
	//---update code frequency and keep table sorted by frequencies
	for(i=0;i<NCODES;i++) {
		if(codefreq[cfctx][i].code==(UINT8)code){
			codefreq[cfctx][i].count++;
			while((i>0) && (codefreq[cfctx][i].count>codefreq[cfctx][i-1].count)){
				tmpcode=codefreq[cfctx][i].code;
				tmpcount=codefreq[cfctx][i].count;
				codefreq[cfctx][i].code=codefreq[cfctx][i-1].code;
				codefreq[cfctx][i].count=codefreq[cfctx][i-1].count;
				codefreq[cfctx][i-1].code=tmpcode;
				codefreq[cfctx][i-1].count=tmpcount;
				i--;
			}
			i=NCODES;
		}
	}
	cfctx=((cfctx<<4)+code)&0x000000FF;
}
//------------------------------------------------------------------------------
// ADD CODE
//------------------------------------------------------------------------------
inline void addcode(UINT32 code)
{
	UINT32 i=0;
	//---search code frequency rank
	while(i<NCODES) {
		if(codefreq[cfctx][i].code==(UINT8)code){
			break;
		}
		i++;
	}
#ifdef VERBOSEMODE
	if(i>(NCODES-1))printf("error in getcode(): code not in range 0..%d\n",NCODES-1);
#endif
	//---add code to queue
	switch (i){
		case 0:
			//binary code 0
			bits++;
			q<<=1;
			break;
		case 1:
			//binary code 10
			bits+=2;
			q<<=2;
			q+=0x0000000000000002ULL;
			break;
		case 2:
			//binary code 110
			bits+=3;
			q<<=3;
			q+=0x0000000000000006ULL;
			break;
		case 3:
			//binary code 1110
			bits+=4;
			q<<=4;
			q+=0x000000000000000EULL;
			break;
		case 4:
			//binary code 11110
			bits+=5;
			q<<=5;
			q+=0x000000000000001EULL;
			break;
		case 5:
			//binary code 11111
			bits+=5;
			q<<=5;
			q+=0x000000000000001FULL;
			break;
	}
	updatecodefreq(code);
}
//------------------------------------------------------------------------------
// UPDATE CACHE
//------------------------------------------------------------------------------
inline void updatecache()
{
	if(matchmode>0)updatebst();

	//---update cache queue
	if(ch!=rank0 && ch!=rank1 && ch!=rank2){
		cachearray[cacheoffset+2]=rank1;
		cachearray[cacheoffset+1]=rank0;
		cachearray[cacheoffset]=(UINT8)ch;
	}else{
		if(ch==rank1){
			cachearray[cacheoffset+1]=rank0;
			cachearray[cacheoffset]=(UINT8)ch;
		}else{
			if(ch==rank2){
				cachearray[cacheoffset+2]=rank1;
				cachearray[cacheoffset+1]=rank0;
				cachearray[cacheoffset]=(UINT8)ch;
			}
		}
	}

	//---update cache context
	switch (contextmode){
		case 0: cachectx=0;break;
		case 1: cachectx=ch; break;
		case 2: cachectx=((cachectx<<8)+ch)&0x0000FFFF;break;
		case 3: cachectx=((cachectx<<8)+ch)&0x00FFFFFF;break;
		case 4: cachectx=((cachectx<<6)+(ch&0x0000003F))&0x00FFFFFF;break;
		case 5: cachectx=((cachectx<<5)+(ch&0x0000001F))&0x00FFFFFF;break;
		case 6: cachectx=((cachectx<<4)+(ch&0x0000000F))&0x00FFFFFF;break;
	}
}
//------------------------------------------------------------------------------
// GET CACHE
//------------------------------------------------------------------------------
inline void getcache()
{
	//---calculate cache offset and get ranks
	cacheoffset=cachectx*3;
	rank0=cachearray[cacheoffset];
	rank1=cachearray[cacheoffset+1];
	rank2=cachearray[cacheoffset+2];

	//---store suffix match search result in context queue
	if(matchmode>0){
		if(matchchar==rank1){
			rank1=rank0;
			rank0=matchchar;
		}else{
			if(matchchar==rank2){
				rank2=rank1;
				rank1=rank0;
				rank0=matchchar;
			}else{
				if(matchchar!=rank0){
					rank2=rank1;
					rank1=rank0;
					rank0=matchchar;
				}
			}
		}
		cachearray[cacheoffset]=rank0;
		cachearray[cacheoffset+1]=rank1;
		cachearray[cacheoffset+2]=rank2;
	}

}
//------------------------------------------------------------------------------
// GET MTF
//------------------------------------------------------------------------------
inline void getmtf()
{
	UINT32 i;
	//---get mtf position
	for(i=0;i<256;i++) {
		if(mtf[mtfctx][i]==ch){
			mtfpos=i;
			break;
		}
	}
}
//------------------------------------------------------------------------------
// UPDATE MTF
//------------------------------------------------------------------------------
inline void updatemtf()
{
	UINT32 i;
	//---move to front
	if(mtfpos!=0){
		mtftarget=(UINT32)(mtfpos>>1);
		mtftemp=mtf[mtfctx][mtfpos];
		for(i=mtfpos;i>mtftarget;i--) {
			mtf[mtfctx][i]=mtf[mtfctx][i-1];
		}
		mtf[mtfctx][mtftarget]=mtftemp;
	}

	//---update MTF context
	mtfctx=((mtfctx<<8)+ch)&0x0000FFFF;
}
//------------------------------------------------------------------------------
// ENCODE
//------------------------------------------------------------------------------
SINT32 encode(char *infile,char *outfile)
{
	clock_t begintime = clock();
	UINT32 i=0;
	UINT32 endoffile=(UINT32)EOF;
	UINT32 outbufptr=0;
	UINT32 runlengthcounter=0;
	UINT8 tmpcharacter=0;
	UINT32 tmpcount=0;
	const SINT32 BLOCKMINUS=2+(QLEN/8);
	const UINT32 BLOCKTRIGGER=(UINT32)(BLOCKLIMIT-BLOCKMINUS);

	//---Pass-1: build symbol frequency table
	if(passmode==2){
		in=fopen(infile,"rb");
		if(in==NULL) {
			fclose(in);
			printf("can't open input file\n");
			return 1;
		}
		for(i=0;i<256;i++) {
			freq[i].character=(UINT8)i;
			freq[i].count=0;
		}
		while(ch=(UINT32)fgetc(in),ch!=endoffile) {
			for(i=0;i<256;i++) { //search ch
				if(freq[i].character==(UINT8)ch){
					freq[i].count++;
					while((i>0) && (freq[i].count>freq[i-1].count)){
						tmpcharacter=freq[i].character;
						tmpcount=freq[i].count;
						freq[i].character=freq[i-1].character;
						freq[i].count=freq[i-1].count;
						freq[i-1].character=tmpcharacter;
						freq[i-1].count=tmpcount;
						i--;
					}
					break;
				}
			}
		}
		fclose(in);
	}

	//---init memory
	initmemory();

	//---open files
	in=fopen(infile,"rb");
	out=fopen(outfile,"wb");
	if(in==NULL || out==NULL) {
		fclose(in);
		fclose(out);
		printf("can't open files\n");
		return 1;
	}

	//---write header
	writeheader(out);

	//---process input file
	while(ch=(UINT32)fgetc(in),ch!=endoffile) {
		inbytes++;

		//---get mtf position
		getmtf();
		//---calculate cache offset and get ranks
		getcache();

		//---build codes
		if(rank0==ch) {
			runlengthcounter++;
			if(runlengthcounter==RUNLENGTHLIMIT){
				//RLE
				addcode(5);
				runlengthcounter=0;
			}
		} else {
			while(runlengthcounter>0){
				//Rank-0
				addcode(0);
				runlengthcounter--;
			}
			if(rank1==ch) {
				//Rank-1
				addcode(1);
			}else{
				if(rank2==ch){
					//Rank-2
					addcode(2);
				}else{
					if(mtfpos<16) {
						//MTF<16
						addcode(3);
						bits+=4;
						q<<=4;
						q+=((UINT64)mtfpos)&0x000000000000000FULL;
					}else{
						//MTF>=16
						addcode(4);
						bits+=8;
						q<<=8;
						q+=((UINT64)mtfpos)&0x00000000000000FFULL;
					}
				}
			}
		}

#ifdef VERBOSEMODE
		if(bits>QLEN){
			printf("error in encode(): queue overflow 1\n");
			return 1;
		}
#endif
		//---update orgbuf with original uncompressed values
		orgbufptr++;
		orgbuf[orgbufptr]=(UINT8)ch;
		//---pull queue bytewise to outbuf
		while(bits>=8){
			outbufptr++;
			if(bits>=8){
				bits-=8;
				outbuf[outbufptr]=(UINT8)((q>>bits)&0x00000000000000FFULL);
			}else{
				bits=0;
				outbuf[outbufptr]=(UINT8)(q&0x00000000000000FFULL);
			}
			if(bits>0){
				q=q&(QMASK>>(QLEN-bits));
			}else{
				q=0;
			}
			//---block?
			if(outbufptr==BLOCKTRIGGER){
				if((outbufptr+BLOCKMINUS)>=orgbufptr){
#ifdef VERBOSEMODE
					if(orgbufptr>(UINT32)BLOCKLIMIT){
						printf("error in encode(): allowed block buffer length exceeded 1\n");
					}
#endif
					//---uncompressed block
					orgbuf[0]=orgbufptr; //block id 0 + 7bits block length
					fwrite(orgbuf,sizeof(UINT8),orgbufptr+1,out);
					outbytes+=orgbufptr+1;

					//---Reset code frequencies
					//---(otherwise decoders code-frequencies are out of sync)
					resetcodefreq();
				}else{
					//---compressed block
					if(bits>(QLEN-13)){ //free space for EOB in queue?
						outbufptr++;
						bits-=8;
						outbuf[outbufptr]=(UINT8)((q>>bits)&0x00000000000000FFULL);
						q=q&(QMASK>>(QLEN-bits));
						outbufptr++;
						bits-=8;
						outbuf[outbufptr]=(UINT8)((q>>bits)&0x00000000000000FFULL);
						q=q&(QMASK>>(QLEN-bits));
					}
					//---add EOB (code 11110 + 8 zeroes) to queue
					addcode(4);
					bits+=8;
					q<<=8;
					//---add remaining queue to outbuf
#ifdef VERBOSEMODE
					if(bits>QLEN){
						printf("error in encode(): queue overflow 2\n");
						return 1;
					}
#endif
					while(bits>0){
						outbufptr++;
						if(bits>=8){
							bits-=8;
							outbuf[outbufptr]=(UINT8)((q>>bits)&0x00000000000000FFULL);
						}else{
							bits=0;
							outbuf[outbufptr]=(UINT8)(q&0x00000000000000FFULL);
						}
						if(bits>0){
							q=q&(QMASK>>(QLEN-bits));
						}else{
							q=0;
						}
					}
					//---write compressed block out
#ifdef VERBOSEMODE
					if(outbufptr>(UINT32)BLOCKLIMIT){
						printf("error in encode(): allowed block buffer length exceeded 2\n");
					}
#endif
					outbuf[0]=(0x80+(UINT8)outbufptr); //block id 1 + 7 bits block length
					fwrite(outbuf,sizeof(UINT8),outbufptr+1,out);
					outbytes+=outbufptr+1;
				}
				//---reset block buffers and queue
				bits=0;
				q=0;
				orgbufptr=0;
				outbufptr=0;
			}//of block
		}

		//---update cache
		updatecache();
		//---move to front and update MTF context
		updatemtf();

	}//eof

	//---write uncompressed rest out
	fputc(0x00,out); //last block indicator
	fputc((orgbufptr&0x0000FF00)>>8,out); //block length
	fputc(orgbufptr&0x000000FF,out); //block length
	for(i=0;i<orgbufptr;i++){
		fputc(orgbuf[i+1],out);
	} 
	outbytes+=orgbufptr+3;

	//---close Files
	fclose(in);
	fclose(out);

	//---add header length
	outbytes+=headerlength;

	//---display result
	printf ("%d Bytes encoded to %d Bytes\n",inbytes,outbytes);
	printf ("Ratio: %f = %f bpb\n",(float)outbytes/inbytes,((float)outbytes/inbytes)*8);
	totaltime=((double)(clock()-begintime))/CLOCKS_PER_SEC;
	printf("Time : %.3f sec\n",totaltime);
	printf("Speed: %.3f Mbyte/s in, %.3f Mbyte/s out\n",(inbytes/totaltime)/(1024*1024),(outbytes/totaltime)/(1024*1024));

	//---end of encoding
	return 0;
}
//------------------------------------------------------------------------------
// DECODE
//------------------------------------------------------------------------------
SINT32 decode(char *infile,char *outfile)
{
	clock_t begintime = clock();
	UINT8 stop=0,stopblock=0;
	UINT8 blockid=0;
	UINT8 blockmode=0;
	UINT32 blockidx=0;
	UINT32 b=0,run=0,r=0;
	SINT32 bctr=0;
	const SINT32 MAXCODELENGTH=(SINT32)(NCODES-1); 

	//---open files
	in=fopen(infile,"rb");
	out=fopen(outfile,"wb");
	if(in==NULL || out==NULL) {
		fclose(in);
		fclose(out);
		printf("can't open files\n");
		return 1;
	}

	//---read header
	if(readheader(in)==1){
		printf("unknown input file format\n");
		return 1;
	}

	//---init memory
	initmemory();

	stop=0;
	while(stop==0){
		blockid=fgetc(in);
		inbytes++;
		if((blockid&0x80)==0x00){
			if(blockid==0x00){
				//---last block
				blockmode=2;
				orgbufptr=(UINT32)((fgetc(in)*256)+fgetc(in)); //16bit block length
				inbytes+=2;
				fread(orgbuf,sizeof(UINT8),orgbufptr,in);
				inbytes+=orgbufptr;
				fwrite(orgbuf,sizeof(UINT8),orgbufptr,out);
				outbytes+=orgbufptr;
				stop=1;
			}else{
				//---uncompressed block
				blockmode=0;
				orgbufptr=(UINT32)(blockid&0x7F); //7bit block length
				fread(orgbuf,sizeof(UINT8),orgbufptr,in);
				inbytes+=orgbufptr;
			}
		}else{
			//---compressed block
			blockmode=1;
			orgbufptr=(UINT32)(blockid&0x7F); //7bit block length
			fread(orgbuf,sizeof(UINT8),orgbufptr,in);
			inbytes+=orgbufptr;
		}

		if(blockmode==0){ //---uncompressed block
			for(blockidx=0;blockidx<orgbufptr;blockidx++){
				ch=orgbuf[blockidx];
				fputc(ch,out);
				outbytes++;
				//---calculate cache offset and get ranks
				getcache();
				//---update cache
				updatecache();
				//---get mtf position
				getmtf();
				//---move to front and update MTF context
				updatemtf();
			}
			//---Reset code frequencies
			//---(Sync with encoder)
			resetcodefreq();
		}else{//---compressed or last block
			if(blockmode==1){//---compressed block
				q=0ULL;
				bits=0;
				blockidx=0;
				stopblock=0;
				while(stopblock==0){
	
					//---calculate cache offset and get ranks
					getcache();

					//---scan bitwise to reconstruct frequency-code	
					bctr=0;
					while(bctr<MAXCODELENGTH){
						if(bits<1){
							q<<=8;
							blockidx++;
							q+=(UINT64)orgbuf[blockidx-1];
							bits+=8;
						}
						if(bits>1){
							b=(UINT32)(q>>(bits-1));
							q=q&(QMASK>>((QLEN+1)-bits));
							bits-=1;
						}else{
							b=(UINT32)q;
							q=0ULL;
							bits-=1;
						}
						if(b==1){
							bctr++;
						}else{
							break;
						}
					}

					bctr=codefreq[cfctx][bctr].code; //decode frequency-code
					updatecodefreq(bctr); //keep code frequencies in sync with encoder
	
					run=1;
					switch (bctr){
						case 0: ch=rank0; break;
						case 1: ch=rank1; break;
						case 2: ch=rank2; break;
						case 3:
							//---read 4 bit MTF-index
							if(bits<4){
								q<<=8;
								blockidx++;
								q+=(UINT64)orgbuf[blockidx-1];
								bits+=8;
							}
							if(bits>4){
								b=(UINT32)(q>>(bits-4));
								q=q&(QMASK>>((QLEN+4)-bits));
								bits-=4;
							}else{
								b=(UINT32)q;
								q=0;
								bits-=4;
							}
							//---rank <16
							ch=mtf[mtfctx][b];
#ifdef VERBOSEMODE
							if(b>=16){
								printf("error in decode(): MTF index >= 16\n");
							}
#endif
							break;
						case 4:
							//---read 8 bit MTF-index or EOB
							if(bits<8){
								q<<=8;
								blockidx++;
								q+=(UINT64)orgbuf[blockidx-1];
								bits+=8;
							}
							if(bits>8){
								b=(UINT32)(q>>(bits-8));
								q=q&(QMASK>>((QLEN+8)-bits));
								bits-=8;
							}else{
								b=(UINT32)q;
								q=0ULL;
								bits-=8;
							}
							if(b!=0){
								//---rank >=16
								ch=mtf[mtfctx][b];
							}else{
								//---EOB
								stopblock=1;
								run=0;
							}
							break;
						case 5:
							//---run of rank-0s
							ch=rank0;
							run=RUNLENGTHLIMIT;
							break;
						case 6:
							//---match
							ch=matchchar;
							break;
					}//of bctr switch
	
	
					for(r=0;r<run;r++){
						fputc((UINT8)ch,out);
						outbytes++;
						//---update cache
						updatecache();
						//---get mtf position
						getmtf();
						//---move to front and update MTF context
						updatemtf();
						//---calculate cache offset and get ranks
						getcache();
						//---load ch
						ch=rank0;
					}//of run	
				}//of stopblock
			}//of blockmode=1
		}//of compressed or last block
	}//of stop

	//---close Files
	fclose(in);
	fclose(out);

	//---add header bytes
	inbytes+=headerlength;

	//---display result
	printf("%d Bytes decoded to %d Bytes\n",inbytes,outbytes);
	printf("Ratio: %f = %f bpb\n",(float)outbytes/inbytes,((float)outbytes/inbytes)*8);
	totaltime=((double)(clock()-begintime))/CLOCKS_PER_SEC;
	printf("Time : %.3f sec\n",totaltime);
	printf("Speed: %.3f Mbyte/s in, %.3f Mbyte/s out\n",(inbytes/totaltime)/(1024*1024),(outbytes/totaltime)/(1024*1024));
	//---end of decoding
	return 0;
}
//------------------------------------------------------------------------------
// USAGE
//------------------------------------------------------------------------------
void usage()
{
	printf("This program is free software; you can redistribute it and/or modify it under");
	printf("\nthe terms of the GNU General Public License.");
	printf("\n");
	printf("\nUsage: symbra [options] <infile> <outfile>");
	printf("\n");
	printf("\nOptions are:");
	printf("\n -c[0-6]  Length of context (default=4)");
	printf("\n -m[0-5]  Suffix match mode (default=0=off). Mode 0-5 uses the following");
	printf("\n          memory in MB, buffer sizes in MB and maximum match lengths:");
	printf("\n          0=65/0/0, 1=74/1/4, 2=83/2/8, 3=92/3/16, 4=101/4/32, 5=110/5/64");
	printf("\n -p[1-2]  Number of passes (default=1)");
	printf("\n -d       Decodes <infile> to <outfile>");
	printf("\n");
	printf("\nEncoding example: symbra -c3 -m3 -p2 myfile.ext myfile.ext.symbra");
	printf("\nDecoding example: symbra -d myfile.ext.symbra myfile.ext");
	printf("\n");
	printf("\nFiles with a size >= 4 GB are not supported.");
	printf("\nPlease report bugs to: frank.schwellinger@gmx.de\n");
}
//------------------------------------------------------------------------------
// MAIN
//------------------------------------------------------------------------------
SINT32 main(SINT32 argc,char *argv[] )
{
	SINT32 i;
	SINT32 codecmode,innum=0,outnum=0;
	FILE *in,*out;
	printf("Symbra %d.%d, a symbol ranking file (de-)compressor, (c) 2007 Frank Schwellinger\n",VERSION,SUBVERSION);
	//---default values
	codecmode=0;
	contextmode=4;
	passmode=1;
	matchmode=0;
	//---scan command line
	if(argc<3) {
		usage();
		exit(1);
	}
	for(i=1;i<argc;i++) {
		char *sz=argv[i];
		if(*sz=='-') {
			switch(*(sz+1)) {
				case 'm':
				case 'M':
					//---Match mode
					switch(*(sz+2)) {
						case '0':
							matchmode=0;
							break;
						case '1':
							matchmode=1;
							break;
						case '2':
							matchmode=2;
							break;
						case '3':
							matchmode=3;
							break;
						case '4':
							matchmode=4;
							break;
						case '5':
							matchmode=5;
							break;
						default:
							printf("unknown argument: allowed is matchmode 0 to 5 \n");
							exit(1);
					}
					break;
				case 'p':
				case 'P':
					//---Number of Passes
					switch(*(sz+2)) {
						case '1':
							passmode=1;
							break;
						case '2':
							passmode=2;
							break;
						default:
							printf("unknown argument: allowed is pass 1 or 2 \n");
							exit(1);
					}
					break;
				case 'c':
				case 'C':
					//---Context mode
					switch(*(sz+2)) {
						case '0':
							contextmode=0;
							break;
						case '1':
							contextmode=1;
							break;
						case '2':
							contextmode=2;
							break;
						case '3':
							contextmode=3;
							break;
						case '4':
							contextmode=4;
							break;
						case '5':
							contextmode=5;
							break;
						case '6':
							contextmode=6;
							break;
						default:
							printf("unknown argument: allowed is context length 0-6\n");
							exit(1);
					}
					break;
				case 'd':
				case 'D':
					//---decode
					if(argc!=4) {
						usage();
						exit(1);
					} else {
						codecmode=1;
						break;
					}
				default:
					printf("unknown argument\n");
					exit(1);
			}
		} else {
			if(argv[i+2]!=NULL) {
				printf("not enough or too many arguments\n");
				exit(1);
			} else {
				innum=i;
				outnum=i+1;
				break;
			}
		}
	}
	//---test input file
	in=fopen(argv[innum],"rb");
	if(in==NULL) {
		printf("can't open input file\n");
		exit(1);
	}
	fclose(in);
	//---test output file
	out=fopen(argv[outnum],"wb");
	if(out==NULL) {
		printf("can't open output file\n");
		exit(1);
	}
	fclose(out);
	//---encode/decode
	if(codecmode==0)
		return encode(argv[innum],argv[outnum]);
	if(codecmode==1)
		return decode(argv[innum],argv[outnum]);
}
//------------------------------------------------------------------------------
// END OF PROGRAM
//------------------------------------------------------------------------------
